/*
 * (C)	Copyright 2001
 * Stubli Faverges - <www.staubli.com>
 * Pierre AUBERT  p.aubert@staubli.com
 * U-Boot port on RPXClassic LF (CLLF_BW31) board
 *
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <config.h>
#include <mpc8xx.h>
#include <spi.h>

/* ------------------------------------------------------------------------- */

static long int dram_size (long int, long int *, long int);
static unsigned char aschex_to_byte (unsigned char *cp);

/* ------------------------------------------------------------------------- */

#define _NOT_USED_	0xFFFFCC05

const uint sdram_table[] =
{
	/*
	 * Single Read. (Offset 00h in UPMA RAM)
	 */
	0xefcbcc04, 0x0f37c804, 0x0eeec004, 0x01b98404,
	0x1ff74c00, /* last */
	_NOT_USED_, _NOT_USED_, _NOT_USED_,

	/*
	 * Burst Read. (Offset 08h in UPMA RAM)
	 */
	0xefcbcc04, 0x0f37c804, 0x0eeec004, 0x00bdc404,
	0x00ffcc00, 0x00ffcc00, 0x01fb8c00, 0x1ff74c00, /* last */
	_NOT_USED_, _NOT_USED_, _NOT_USED_, _NOT_USED_,
	_NOT_USED_, _NOT_USED_, _NOT_USED_, _NOT_USED_,

	/*
	 * Single Write. (Offset 18h in UPMA RAM)
	 */
	0xefcbcc04, 0x0f37c804, 0x0eee8002, 0x01b90404,
	0x1ff74c05, /* last */
	_NOT_USED_, _NOT_USED_, _NOT_USED_,

	/*
	 * Burst Write. (Offset 20h in UPMA RAM)
	 */
	0xefcbcc04, 0x0f37c804, 0x0eee8000, 0x00bd4400,
	0x00ffcc00, 0x00ffcc02, 0x01fb8c04, 0x1ff74c05, /* last */
	_NOT_USED_, _NOT_USED_, _NOT_USED_, _NOT_USED_,
	_NOT_USED_, _NOT_USED_, _NOT_USED_, _NOT_USED_,

	/*
	 * Refresh. (Offset 30h in UPMA RAM)
	 */
	0xeffacc04, 0x0ff5cc04, 0x0fffcc04, 0x1fffcc04,
	0xffffcc05, 0xffffcc05, 0xeffb8c34, 0x0ff74c34, /* last */
	0x0ffaccb4, 0x0ff5cc34, 0x0fffc034, 0x0fffc0b4,

	/*
	 * Exception. (Offset 3Ch in UPMA RAM)
	 */
	0x0fea8034, 0x1fb54034, 0xffffcc34, _NOT_USED_
};

/* ------------------------------------------------------------------------- */


/*
 * Check Board Identity:
 */

int checkboard (void)
{
	puts ("Board: EP866\n");
	return (0);
}

void ep852_init (void)
{
	/* Enable NVRAM */
	*((uchar *) BCSR0) |= BCSR0_ENNVRAM;

	/* Enable the on-board RS232 transiever for the console port */
	*((uchar *) BCSR0) |= BCSR0_ENMONXCVR; 

#ifdef CONFIG_FEC_ENET
	/* Validate the fast ethernet tranceiver */
	*((volatile uchar *) BCSR2) &= ~BCSR2_MIICTL;
	*((volatile uchar *) BCSR2) &= ~BCSR2_MIIPWRDWN;
	*((volatile uchar *) BCSR2) |= BCSR2_MIIRST;
	*((volatile uchar *) BCSR2) |= BCSR2_MIIPWRDWN;
#endif

}

/* ------------------------------------------------------------------------- */

long int initdram (int board_type)
{
	volatile immap_t *immap = (immap_t *) CFG_IMMR;
	volatile memctl8xx_t *memctl = &immap->im_memctl;
	long int size;

	upmconfig (UPMA, (uint *) sdram_table,
			   sizeof (sdram_table) / sizeof (uint));

	/* Refresh clock prescalar */
	memctl->memc_mptpr = CFG_MPTPR;

	memctl->memc_mar = 0x00000088;

	/* Map controller banks 1 to the SDRAM bank */
	memctl->memc_or1 = CFG_OR1_PRELIM;
	memctl->memc_br1 = CFG_BR1_PRELIM;

	memctl->memc_mamr = CFG_MAMR_9COL & (~(MAMR_PTAE));	/* no refresh yet */

	udelay (200);

	/* perform SDRAM initialization sequence */

	memctl->memc_mcr = 0x80002236;	/* SDRAM bank 0 - refresh twice */
	udelay (1);

	memctl->memc_mamr |= MAMR_PTAE; /* enable refresh */

	udelay (1000);

	/* Check Bank 0 Memory Size
	 * try 9 column mode
	 */

	size = dram_size (CFG_MAMR_9COL, (ulong *) SDRAM_BASE_PRELIM,
						SDRAM_MAX_SIZE);

	return (size);
}

/* ------------------------------------------------------------------------- */

/*
 * Check memory range for valid RAM. A simple memory test determines
 * the actually available RAM size between addresses `base' and
 * `base + maxsize'. Some (not all) hardware errors are detected:
 * - short between address lines
 * - short between data lines
 */

static long int dram_size (long int mamr_value, long int *base, long int maxsize)
{
	volatile immap_t *immap = (immap_t *) CFG_IMMR;
	volatile memctl8xx_t *memctl = &immap->im_memctl;

	memctl->memc_mamr = mamr_value;

	return (get_ram_size(base, maxsize));
}



/*-------------------- SPI CHIP SELECT CODE -----------------------------
 *
 * The following are used to control the SPI chip selects for the SPI command.
 *
 * The EP852 board uses its CPLD to multiplex the CS for SPI. This multiplex
 * function is controlled by bits 6 & 7 on BSCR4. 
 *
 *    SPI Device         |  cs#  |  bit 6  |  bit 7 
 *   -----------------------------------------------
 *    SEP (EEPROM)       |   0   |    0    |    0
 *    STTM (Temperature) |   1   |    0    |    1
 *    [on expansion bus] |   2   |    1    |    0
 *    [on expansion bus] |   3   |    1    |    1
 * 
 *    Bit 6 is SPI_CS_SEL0 and bit7 is SPI_CS_SEL1 on BSCR4
 */

#if (CONFIG_COMMANDS & CFG_CMD_SPI) || defined(CONFIG_SOFT_SPI)

#if 0
void chipsel(int cs)
{
	volatile immap_t *immr;
	volatile cpm8xx_t *cp;
	immr = (immap_t *)  CFG_IMMR;
	cp   = (cpm8xx_t *) &immr->im_cpm;

	if (cs) /* CS is active, set low */
	cp->cp_pbdat &= ~0x0001;
	else /* CS is inactive, set high */
	cp->cp_pbdat |= 0x0001;
}
#endif

void spi_eeprom_chipsel(int cs)
{	/* This is CS0. It is the chip select for the SPI EEPROM on the board
	   The EEPROM chip is the 25LC040 from Microchip. 
	   The board should boot with this CS selected and inactive (high) as 
	   controlled by the CPLD and SPI initialization. */
	volatile immap_t *immr = (immap_t *)  CFG_IMMR;
	volatile cpm8xx_t *cp = (cpm8xx_t *) &immr->im_cpm;
	
	if (cs) /* cs is active, so set it low */
	{  /* set the SPI control register in the BCSR for CS0 (Switch the 
	   multiplexer on the CPLD to CS0) */
		*((volatile uchar *) BCSR4) &= ~(BCSR4_SPI_CS_1 || BCSR4_SPI_CS_0);
	   /* Now set cs low */
		cp->cp_pbdat &= ~0x0001;
	}
	else    /* CS is inactive, set high */
	{
		cp->cp_pbdat |= 0x0001;
	}	
}

void spi_thermal_chipsel(int cs)
{	/* This is CS 1. It is the chip select for the temperature sensor and 
	   thermal monitor on the board. */
	volatile immap_t *immr = (immap_t *)  CFG_IMMR;
	volatile cpm8xx_t *cp = (cpm8xx_t *) &immr->im_cpm;
	
	if (cs) /* cs is active, so set it low */
	{  /* set the SPI control register in the BCSR for CS1 (Switch the 
	   multiplexer on the CPLD to CS1) */
		*((volatile uchar *) BCSR4) = 
	 	 (BCSR4 & ~BCSR4_SPI_CS_0) || BCSR4_SPI_CS_1;
	   /* Now set cs low */
		cp->cp_pbdat &= ~0x0001;
	}
	else    /* CS is inactive, set high */
	{
		cp->cp_pbdat |= 0x0001;
	}
}

void spi_CS2_chipsel(int cs)
{	/* CS2 is available on the expansion bus */

	volatile immap_t *immr = (immap_t *)  CFG_IMMR;
	volatile cpm8xx_t *cp = (cpm8xx_t *) &immr->im_cpm;
	
	if (cs) /* cs is active, so set it low */
	{  /* set the SPI control register in the BCSR for CS2 (Switch the 
	   multiplexer on the CPLD to CS2) */
		*((volatile uchar *) BCSR4) = 
	  	 BCSR4_SPI_CS_0 || (BCSR4 & ~BCSR4_SPI_CS_1);
	   /* Now set cs low */
		cp->cp_pbdat &= ~0x0001;
	}
	else    /* CS is inactive, set high */
	{
		cp->cp_pbdat |= 0x0001;
	}
}

void spi_CS3_chipsel(int cs)
{	/* CS3 is available on the expansion bus */
	
	volatile immap_t *immr = (immap_t *)  CFG_IMMR;
	volatile cpm8xx_t *cp = (cpm8xx_t *) &immr->im_cpm;
	
	if (cs) /* cs is active, so set it low */
	{  /* set the SPI control register in the BCSR for CS3 (Switch the 
	   multiplexer on the CPLD to CS3) */
	*((volatile uchar *) BCSR4) |= (BCSR4_SPI_CS_1 || BCSR4_SPI_CS_0);
	   /* Now set cs low */
		cp->cp_pbdat &= ~0x0001;
	}
	else    /* CS is inactive, set high */
	{
		cp->cp_pbdat |= 0x0001;
	}
}

/*
 * The SPI command uses this table of functions for controlling the SPI
 * chip selects: it calls the appropriate function to control the SPI
 * chip selects. */
 
spi_chipsel_type spi_chipsel[] = {
	spi_eeprom_chipsel,
	spi_thermal_chipsel,
	spi_CS2_chipsel,
	spi_CS3_chipsel
	};

int spi_chipsel_cnt = sizeof(spi_chipsel) / sizeof(spi_chipsel[0]);


#endif /* CFG_CMD_SPI */
